//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2023 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include <stdafx.h>
#include <at/atcore/enumparseimpl.h>
#include <at/atcore/savestate.h>
#include "tracefileformat.h"

AT_DEFINE_ENUM_TABLE_BEGIN(ATSavedTraceCPUColumnType)
	{ ATSavedTraceCPUColumnType::None,				"none" },
	{ ATSavedTraceCPUColumnType::A,					"a" },
	{ ATSavedTraceCPUColumnType::X,					"x" },
	{ ATSavedTraceCPUColumnType::Y,					"y" },
	{ ATSavedTraceCPUColumnType::S,					"s" },
	{ ATSavedTraceCPUColumnType::P,					"p" },
	{ ATSavedTraceCPUColumnType::PC,				"pc" },
	{ ATSavedTraceCPUColumnType::Opcode,			"opcode" },
	{ ATSavedTraceCPUColumnType::Cycle,				"cycle" },
	{ ATSavedTraceCPUColumnType::UnhaltedCycle,		"unhalted_cycle" },
	{ ATSavedTraceCPUColumnType::Irq,				"irq" },
	{ ATSavedTraceCPUColumnType::Nmi,				"nmi" },
	{ ATSavedTraceCPUColumnType::EffectiveAddress,	"effective_address" },
	{ ATSavedTraceCPUColumnType::GlobalPCBase,		"global_pc_base" },
AT_DEFINE_ENUM_TABLE_END(ATSavedTraceCPUColumnType, ATSavedTraceCPUColumnType::None)

////////////////////////////////////////////////////////////////////////////////

template<ATExchanger T>
void ATSavedTraceRoot::Exchange(T& rw) {
	rw.Transfer("groups", &mGroups);
}

////////////////////////////////////////////////////////////////////////////////

template<ATExchanger T>
void ATSavedTraceGroup::Exchange(T& rw) {
	rw.Transfer("name", &mName);
	rw.Transfer("channels", &mChannels);
}

////////////////////////////////////////////////////////////////////////////////

template<ATExchanger T>
void ATSavedTraceChannel::Exchange(T& rw) {
	rw.Transfer("name", &mName);
	rw.Transfer("detail", &mpDetail);
}

////////////////////////////////////////////////////////////////////////////////

template<typename T_Self, ATExchanger T_Ex>
void ATSavedTraceFrameChannelDetail::FrameBoundaryInfo::Exchange(T_Self& self, T_Ex& rw) {
	rw.Transfer("offset", &self.mOffset);
	rw.Transfer("period", &self.mPeriod);
	rw.Transfer("count", &self.mCount);

	if constexpr(rw.IsReader) {
		if (self.mCount == 0 || self.mPeriod == 0)
			throw ATInvalidSaveStateException();
	}
}

template<ATExchanger T>
void ATSavedTraceFrameChannelDetail::Exchange(T& rw) {
	rw.Transfer("ticks_per_second", &mTicksPerSecond);
	rw.Transfer("frame_boundaries", &mFrameBoundaries);

	if constexpr(rw.IsReader) { 
		if (mTicksPerSecond < 1770000.0 || mTicksPerSecond > 1790000.0)
			throw ATUnsupportedSaveStateException();
	}
}

////////////////////////////////////////////////////////////////////////////////

template<typename T_Self, ATExchanger T_Ex>
void ATSavedTraceCPUChannelDetail::ColumnInfo::Exchange(T_Self& self, T_Ex& rw) {
	rw.Transfer("type", &self.mType);
	rw.Transfer("bit_offset", &self.mBitOffset);
	rw.Transfer("bit_width", &self.mBitWidth);
}

template<typename T_Self, ATExchanger T_Ex>
void ATSavedTraceCPUChannelDetail::StripeInfo::Exchange(T_Self& self, T_Ex& rw) {
	rw.Transfer("codec", &self.mpCodec);
	rw.Transfer("columns", &self.mColumns);
	rw.Transfer("byte_count", &self.mByteCount);
	rw.Transfer("blocks", &self.mBlocks);
}

template<ATExchanger T>
void ATSavedTraceCPUChannelDetail::Exchange(T& rw) {
	rw.Transfer("row_count", &mRowCount);
	rw.Transfer("row_group_size", &mRowGroupSize);
	rw.Transfer("column_stripes", &mColumnStripes);
	rw.Transfer("predictors", &mPredictors);
}

////////////////////////////////////////////////////////////////////////////////

template<ATExchanger T>
void ATSavedTraceEventChannelDetail::Exchange(T& rw) {
	rw.Transfer("string_table", &mStringTable);
	rw.Transfer("color_table", &mColorTable);
	rw.Transfer("event_buffer", &mpEventBuffer);
}

////////////////////////////////////////////////////////////////////////////////

template<ATExchanger T>
void ATSavedTraceVideoChannelDetail::Exchange(T& rw) {
	rw.Transfer("frame_count", &mFrameCount);
	rw.Transfer("frame_buffer_count", &mFrameBufferCount);
	rw.Transfer("frame_buffers_per_group", &mFrameBuffersPerGroup);
	rw.Transfer("frame_info", &mpFrameInfo);
	rw.Transfer("frame_buffer_info", &mpFrameBufferInfo);
	rw.Transfer("frame_buffer_groups", &mFrameBufferGroups);

	if constexpr(rw.IsReader) {
		if (!mFrameBuffersPerGroup || mFrameBuffersPerGroup > 65536)
			throw ATInvalidSaveStateException();
	}
}
